/**
 * @file raw_udp6.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <arpa/inet.h>
#include <linux/filter.h>
#include <net/ethernet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>

#include "raw_udp6.h"
#include "config.h"
#include "contain.h"
#include "ether.h"
#include "print.h"
#include "transport_private.h"
#include "util.h"
#include "raw_udp6_bpf_api.h"
#include "raw_udp_common.h"

#define EVENT_PORT        319
#define GENERAL_PORT      320

/* The 0x0e in second byte is substituted with udp6_scope in runtime. */
#define PTP_PRIMARY_MCAST_IP6ADDR "FF0E:0:0:0:0:0:0:181"

struct udp_hdr {
	uint16_t udp_src;
	uint16_t udp_dst;
	uint16_t udp_len;
	uint16_t udp_check;
} __attribute__((packed));

struct ip6_hdr {
	union {
		struct ip6_hdrctl {
			uint32_t ip6_un1_flow;/* 20 bits of flow-ID */
			uint16_t ip6_un1_plen;/* payload length */
			uint8_t  ip6_un1_nxt;/* next header */
			uint8_t  ip6_un1_hlim;/* hop limit */
		} ip6_un1;
		u_int8_t ip6_un2_vfc;/* 4 bits version, top 4 bits class */
	} ip6_ctlun;
	struct in6_addr ip6_src;/* source address */
	struct in6_addr ip6_dst;/* destination address */
} __attribute__((packed));

struct pseudo6_hdr {
	struct in6_addr ip6_src;/* source address */
	struct in6_addr ip6_dst;/* destination address */
	uint32_t length;
	uint16_t reserved1;
	uint8_t reserved2;
	uint8_t next_header;
} __attribute__((packed));

struct raw_udp6 {
	struct transport t;

	/* IPv6 */
	struct in6_addr src_ip;
	struct in6_addr primary_ip;
	unsigned dscp_general;
	unsigned dscp_event;
	int hlim;

	/* MAC */
	struct address src_addr;
	eth_addr ptp_addr;
	eth_addr p2p_addr;
	unsigned vlan;
};

static int raw_udp6_close(struct transport *t, struct fdarray *fda)
{
	close(fda->fd[0]);
	close(fda->fd[1]);
	if (fda->sock_event > 0) {
		close(fda->sock_event);
	}
	if (fda->sock_general > 0) {
		close(fda->sock_general);
	}
	return 0;
}

#define ARRAY_SIZE(array) (sizeof(array) / sizeof(*array))

static struct sock_filter raw_filter[] = {
#include "raw_udp6_bpf.h"
};

const struct raw_two_insns {
	struct sock_filter insns[2];
} raw_st_mem_template = {{
	{ 0x00,  0,  0, 0x00000000 }, /* ldi 0 */
	{ 0x02,  0,  0, 0x00000000 }, /* st M[0] */
}};

struct raw_two_insns jit_st_mem(int value, int address)
{
	struct raw_two_insns insns = raw_st_mem_template;
	insns.insns[0].k = value;
	insns.insns[1].k = address;
	return insns;
}

struct raw_two_insns jit_pass_port(int port)
{
	return jit_st_mem(port, RAW_UDP6_M_PORT);
}

struct raw_two_insns jit_pass_vlan(int vlan)
{
	return jit_st_mem(vlan, RAW_UDP6_M_VLAN);
}

void jit_append(struct sock_fprog *prog,
		const struct sock_filter *insns,
		unsigned short length)
{
	memcpy(prog->filter + prog->len, insns, length * sizeof(*insns));
	prog->len += length;
}

void jit_append_two(struct sock_fprog *prog,
		    const struct raw_two_insns insns)
{
	jit_append(prog, insns.insns, ARRAY_SIZE(insns.insns));
}

struct sock_fprog jit_build(int port, int vlan,
			    struct sock_filter *insns, size_t ninsns)
{
	struct sock_fprog prg = {0, insns};

	jit_append_two(&prg, jit_pass_port(port));
	jit_append_two(&prg, jit_pass_vlan(vlan));
	jit_append(&prg, raw_filter, ARRAY_SIZE(raw_filter));
	if (prg.len > ninsns)
		abort();
	return prg;
}

static void raw_udp6_mc_ip_2_mac(struct in6_addr in6_addr,
				 unsigned char mac[MAC_LEN])
{
	mac[0] = 0x33;
	mac[1] = 0x33;
	mac[2] = in6_addr.s6_addr[12];
	mac[3] = in6_addr.s6_addr[13];
	mac[4] = in6_addr.s6_addr[14];
	mac[5] = in6_addr.s6_addr[15];
}

static int raw_udp6_open(struct transport *t, struct interface *iface,
			struct fdarray *fda, enum timestamp_type ts_type)
{
	struct raw_udp6 *raw = container_of(t, struct raw_udp6, t);
	struct sock_filter insns[ARRAY_SIZE(raw_filter) + 4];
	struct sock_fprog prg;
	unsigned char ptp_dst_mac[MAC_LEN];
	unsigned char p2p_dst_mac[MAC_LEN];
	int efd, gfd;
	char *str;
	char *vlan_intf = config_get_string(t->cfg, interface_name(iface), "vlan_intf");
	char *lag_intf = config_get_string(t->cfg, interface_name(iface), "lag_intf");
	char *netdev;
	char *intf;
	int sock_event, sock_general;

	netdev = alloca(strlen(interface_name(iface)) + 1);
	strcpy(netdev, interface_name(iface));
	*strchrnul(netdev, '@') = '\0';

	if (strlen(vlan_intf))
		intf = vlan_intf;
	else if (strlen(lag_intf))
		intf = lag_intf;
	else
		intf = netdev;

	raw->vlan = config_get_int(t->cfg, interface_name(iface), "vlan");
	raw->dscp_event = config_get_int(t->cfg, interface_name(iface), "dscp_event");
	raw->dscp_general = config_get_int(t->cfg, interface_name(iface), "dscp_general");
	raw->hlim = config_get_int(t->cfg, interface_name(iface), "udp_ttl");

	str = config_get_string(t->cfg, interface_name(iface), "p2p_dst_mac");
	if (str2mac(str, p2p_dst_mac)) {
		pr_err("invalid p2p_dst_mac %s", str);
		return -1;
	}

	str = config_get_string(t->cfg, interface_name(iface), "src_ip");
	if (inet_pton(AF_INET6, str, &raw->src_ip) != 1) {
		pr_err("invalid src_ip %s", str);
		return -1;
	}

	if (inet_pton(AF_INET6, PTP_PRIMARY_MCAST_IP6ADDR,
		      &raw->primary_ip) != 1)
		return -1;
	raw->primary_ip.s6_addr[1] = config_get_int(t->cfg, interface_name(iface),
						    "udp6_scope");
	raw_udp6_mc_ip_2_mac(raw->primary_ip, ptp_dst_mac);

	memcpy(raw->ptp_addr, ptp_dst_mac, MAC_LEN);
	memcpy(raw->p2p_addr, p2p_dst_mac, MAC_LEN);
#ifdef NAME_SPACE
	if(!switch_namespace(interface_namespace(iface))) {
		exit(-1);
	}
#endif
	if (sk_interface_macaddr(intf, &raw->src_addr))
		goto no_mac;

#ifdef NAME_SPACE
	if (strlen(intf)) {
		if(!switch_namespace("")) {
			exit(-1);
		}
	}
#endif

	sock_event = raw_udp_common_bind_port_to_socket(EVENT_PORT, 0);
	if (sock_event < 0) {
		pr_warning("failed to bind event port to socket");
	}

	sock_general = raw_udp_common_bind_port_to_socket(GENERAL_PORT, 0);
	if (sock_general < 0) {
		pr_warning("failed to bind general port to socket");
	}

	prg = jit_build(EVENT_PORT, raw->vlan, insns, ARRAY_SIZE(insns));
	efd = raw_udp_common_open_socket(netdev, 1,
					 ptp_dst_mac, p2p_dst_mac, &prg);
	if (efd < 0)
		goto no_event;

	prg = jit_build(GENERAL_PORT, raw->vlan, insns, ARRAY_SIZE(insns));
	gfd = raw_udp_common_open_socket(netdev, 0,
					 ptp_dst_mac, p2p_dst_mac, &prg);
	if (gfd < 0)
		goto no_general;

	if (sk_timestamping_init(efd, netdev, ts_type, TRANS_RAW_UDP_IPV6))
		goto no_timestamping;

	if (sk_general_init(gfd))
		goto no_timestamping;

	fda->fd[FD_EVENT] = efd;
	fda->fd[FD_GENERAL] = gfd;
	fda->sock_event = sock_event;
	fda->sock_general = sock_general;
	return 0;

no_timestamping:
	close(gfd);
no_general:
	close(efd);
no_event:
no_mac:
	return -1;
}

static int raw_udp6_recv(struct transport *t, int fd, void *buf, int buflen,
			 struct address *addr, struct hw_timestamp *hwts)
{
	int cnt, hlen;
	unsigned char *ptr = buf;
	struct ip6_hdr *ip6_hdr = NULL;

	/* This may have to skip IPv6 options if they are accepted by the
	 * filter. */
	hlen = sizeof(struct udp_hdr) + sizeof(struct ip6_hdr) +
		sizeof(struct eth_hdr);

	/* Shifts into reserved headroom in struct message_storage. */
	ptr    -= hlen;
	buflen += hlen;

	cnt = sk_receive(fd, ptr, buflen, addr, hwts, 0);

	ip6_hdr = (struct ip6_hdr *) (ptr + sizeof(struct eth_hdr));
	addr->peer_ip.sin6.sin6_family = AF_INET6;
	memcpy(addr->peer_ip.sin6.sin6_addr.s6_addr,
	       ip6_hdr->ip6_src.s6_addr,
	       sizeof(ip6_hdr->ip6_src.s6_addr));

	if (cnt < 0)
		return cnt;

	cnt -= hlen;
	return cnt;
}

static uint16_t udp6_checksum(struct ip6_hdr *ip6_hdr, void *buf,
			      int udp_length)
{
	unsigned char pkt[1600] = {0};
	struct pseudo6_hdr ps6_hdr = {0};

	ps6_hdr.ip6_src = ip6_hdr->ip6_src;
	ps6_hdr.ip6_dst = ip6_hdr->ip6_dst;
	ps6_hdr.length = htons(udp_length);
	ps6_hdr.next_header = IPPROTO_UDP;

	memcpy(pkt, &ps6_hdr, sizeof(ps6_hdr));
	memcpy(pkt + sizeof(ps6_hdr), buf, udp_length);

	return raw_udp_common_checksum(pkt, sizeof(ps6_hdr) + udp_length);
}

static int raw_udp6_send(struct transport *t, struct fdarray *fda, 
	                 enum transport_event event, int peer, void *buf, 
			 int len, struct address *addr, struct hw_timestamp *hwts)
{
	struct raw_udp6 *raw = container_of(t, struct raw_udp6, t);
	struct ptp_message *msg = buf;
	ssize_t cnt;
	uint8_t dscp = event ? raw->dscp_event : raw->dscp_general;
	int fd = event ? fda->fd[FD_EVENT] : fda->fd[FD_GENERAL];
	unsigned char pkt[1600], *ptr = buf;
	struct vlan_hdr *vlan_hdr;
	struct eth_hdr *eth_hdr;
	struct ip6_hdr *ip6_hdr;
	struct udp_hdr *udp_hdr;
	uint8_t *ethaddr;

	/* UDP Header */
	ptr -= sizeof(*udp_hdr);
	len += sizeof(*udp_hdr);

	udp_hdr = (struct udp_hdr *) ptr;
	udp_hdr->udp_src = event ? htons(EVENT_PORT) : htons(GENERAL_PORT);
	udp_hdr->udp_dst = event ? htons(EVENT_PORT) : htons(GENERAL_PORT);
	udp_hdr->udp_len = htons(len);

	/* IPv6 Header */
	ptr -= sizeof(*ip6_hdr);
	ip6_hdr = (struct ip6_hdr *) ptr;

	ip6_hdr->ip6_src = raw->src_ip;
	if (msg->header.flagField[0] & UNICAST)
		ip6_hdr->ip6_dst = msg->address.peer_ip.sin6.sin6_addr;
	else
		ip6_hdr->ip6_dst = raw->primary_ip;
	ip6_hdr->ip6_ctlun.ip6_un1.ip6_un1_hlim = raw->hlim;

	udp_hdr->udp_check = 0;
	udp_hdr->udp_check = udp6_checksum(ip6_hdr,
					   ptr + sizeof(*ip6_hdr), len);

	ip6_hdr->ip6_ctlun.ip6_un1.ip6_un1_flow = htonl(dscp << 22);
	ip6_hdr->ip6_ctlun.ip6_un1.ip6_un1_plen = htons(len);
	ip6_hdr->ip6_ctlun.ip6_un1.ip6_un1_nxt = IPPROTO_UDP; /* UDP */
	ip6_hdr->ip6_ctlun.ip6_un2_vfc |= 6 << 4;
	len += sizeof(*ip6_hdr);

	if (addr)
		ethaddr = &addr->sll.sll_addr[0];
	else
		ethaddr = peer ? &raw->p2p_addr[0] : &raw->ptp_addr[0];

	if (raw->vlan != 0) {
		/* Ethernet VLAN Header */
		ptr -= sizeof(*vlan_hdr);
		len += sizeof(*vlan_hdr);
		vlan_hdr = (struct vlan_hdr *) ptr;

		memcpy(vlan_hdr->dst, ethaddr, sizeof(vlan_hdr->dst));
		memcpy(vlan_hdr->src, raw->src_addr.sll.sll_addr,
		       sizeof(vlan_hdr->src));
		vlan_hdr->tpid = htons(ETH_P_8021Q);
		vlan_hdr->tci = htons((raw->vlan | 0xC000));
		vlan_hdr->type = htons(ETH_P_IPV6);
	} else {
		ptr -= sizeof(*eth_hdr);
		len += sizeof(*eth_hdr);
		eth_hdr = (struct eth_hdr *) ptr;

		memcpy(eth_hdr->dst, ethaddr, sizeof(eth_hdr->dst));
		memcpy(eth_hdr->src, raw->src_addr.sll.sll_addr,
		       sizeof(eth_hdr->src));
		eth_hdr->type = htons(ETH_P_IPV6);
	}

	cnt = send(fd, ptr, len, 0);
	if (cnt < 1) {
		pr_err("send failed: %d %m", errno);
		return cnt;
	}
	/*
	 * Get the time stamp right away.
	 */
	return event == TRANS_EVENT ?
		sk_receive(fd, pkt, len, NULL, hwts, MSG_ERRQUEUE) : cnt;
}

static int raw_udp6_physical_addr(struct transport *t, uint8_t *addr)
{
	struct raw_udp6 *raw = container_of(t, struct raw_udp6, t);
	memcpy(addr, raw->src_addr.sll.sll_addr, MAC_LEN);
	return MAC_LEN;
}

static int raw_udp6_protocol_addr(struct transport *t, uint8_t *addr)
{
	struct raw_udp6 *raw = container_of(t, struct raw_udp6, t);
	memcpy(addr, raw->src_addr.sll.sll_addr, MAC_LEN);
	return MAC_LEN;
}

static void raw_udp6_release(struct transport *t)
{
	struct raw_udp6 *raw = container_of(t, struct raw_udp6, t);
	free(raw);
}

struct transport *raw_udp6_transport_create(void)
{
	struct raw_udp6 *raw;
	raw = calloc(1, sizeof(*raw));
	if (!raw)
		return NULL;
	raw->t.close   = raw_udp6_close;
	raw->t.open    = raw_udp6_open;
	raw->t.recv    = raw_udp6_recv;
	raw->t.send    = raw_udp6_send;
	raw->t.release = raw_udp6_release;
	raw->t.physical_addr = raw_udp6_physical_addr;
	raw->t.protocol_addr = raw_udp6_protocol_addr;
	return &raw->t;
}
