/**
 * @file raw_udp_common.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <arpa/inet.h>
#include <net/ethernet.h>
#include <stddef.h>
#include <string.h>
#include <unistd.h>

#include "raw_udp_common.h"
#include "print.h"

static int raw_udp_common_configure(int fd, int event, int index,
				    unsigned char *addr1, unsigned char *addr2,
				    int enable, struct sock_fprog *prg)
{
	int err1, err2, option;
	struct packet_mreq mreq;

	if (setsockopt(fd, SOL_SOCKET, SO_ATTACH_FILTER, prg, sizeof(*prg))) {
		pr_err("setsockopt SO_ATTACH_FILTER failed: %m");
		return -1;
	}

	option = enable ? PACKET_ADD_MEMBERSHIP : PACKET_DROP_MEMBERSHIP;

	memset(&mreq, 0, sizeof(mreq));
	mreq.mr_ifindex = index;
	mreq.mr_type = PACKET_MR_MULTICAST;
	mreq.mr_alen = MAC_LEN;
	memcpy(mreq.mr_address, addr1, MAC_LEN);

	err1 = setsockopt(fd, SOL_PACKET, option, &mreq, sizeof(mreq));
	if (err1)
		pr_warning("setsockopt PACKET_MR_MULTICAST failed: %m");

	memcpy(mreq.mr_address, addr2, MAC_LEN);

	err2 = setsockopt(fd, SOL_PACKET, option, &mreq, sizeof(mreq));
	if (err2)
		pr_warning("setsockopt PACKET_MR_MULTICAST failed: %m");

	if (!err1 && !err2)
		return 0;

	mreq.mr_ifindex = index;
	mreq.mr_type = PACKET_MR_ALLMULTI;
	mreq.mr_alen = 0;
	if (!setsockopt(fd, SOL_PACKET, option, &mreq, sizeof(mreq))) {
		return 0;
	}
	pr_warning("setsockopt PACKET_MR_ALLMULTI failed: %m");

	mreq.mr_ifindex = index;
	mreq.mr_type = PACKET_MR_PROMISC;
	mreq.mr_alen = 0;
	if (!setsockopt(fd, SOL_PACKET, option, &mreq, sizeof(mreq))) {
		return 0;
	}
	pr_warning("setsockopt PACKET_MR_PROMISC failed: %m");

	pr_err("all socket options failed");
	return -1;
}

int raw_udp_common_bind_port_to_socket(short port, short is_ipv4)
{
    int sock;
    struct sockaddr_in addr;
    struct sockaddr_in6 addr6;
    int on = 1;

    sock = socket(is_ipv4 ? PF_INET : PF_INET6, SOCK_DGRAM, IPPROTO_UDP);
    if (sock < 0) {
        pr_err("socket failed: %m");
        goto no_socket;
    }
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))) {
        pr_err("setsockopt SO_REUSEADDR failed: %m");
        goto no_option;
    }

    if (is_ipv4) {
        memset(&addr, 0, sizeof(addr));
        addr.sin_family = AF_INET;
        addr.sin_port = htons(port);

        if (bind(sock, (struct sockaddr_in *) &addr, sizeof(addr))) {
            pr_err("bind failed: %m");
            goto no_option;
        }
    } else {
        memset(&addr6, 0, sizeof(addr6));
        addr6.sin6_family = AF_INET6;
        addr6.sin6_port = htons(port);

        if (bind(sock, (struct sockaddr_in6 *) &addr6, sizeof(addr6))) {
            pr_err("bind failed: %m");
            goto no_option;
        }
    }

    return sock;
no_option:
    close(sock);
no_socket:
    return -1;
}

int raw_udp_common_open_socket(const char *name, int event,
			       unsigned char *ptp_dst_mac,
			       unsigned char *p2p_dst_mac,
			       struct sock_fprog *prg)
{
	struct sockaddr_ll addr;
	int fd, index;
	const uint32_t rmem_default = 6291456;
	int priority = 6;

	fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
	if (fd < 0) {
		pr_err("socket failed: %m");
		goto no_socket;
	}
	index = sk_interface_index(fd, name);
	if (index < 0)
		goto no_option;

	memset(&addr, 0, sizeof(addr));
	addr.sll_ifindex = index;
	addr.sll_family = AF_PACKET;
	addr.sll_protocol = htons(ETH_P_ALL);
	if (bind(fd, (struct sockaddr *) &addr, sizeof(addr))) {
		pr_err("bind failed: %m");
		goto no_option;
	}
	if (setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, name, strlen(name))) {
		pr_err("setsockopt SO_BINDTODEVICE failed: %m");
		goto no_option;
	}
	if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &rmem_default, sizeof(uint32_t))) {
		pr_warning("setsockopt SO_RCVBUF failed: %m");
	}
	if(setsockopt(fd, SOL_SOCKET, SO_PRIORITY, &priority, sizeof(priority))){
		pr_warning("setsockopt SO_PRIORITY failed: %m");
	}

	if (raw_udp_common_configure(fd, event, index,
				     ptp_dst_mac, p2p_dst_mac, 1, prg))
		goto no_option;

	return fd;
no_option:
	close(fd);
no_socket:
	return -1;
}

/* http://www.microhowto.info/howto/calculate_an_internet_protocol_checksum_in_c.html */
uint16_t raw_udp_common_checksum(void *buf, int length)
{
	char* data=(char*)buf;

	uint64_t acc=0xffff;

	unsigned int offset=((uintptr_t)data)&3;
	if (offset) {
		size_t count=4-offset;
		if (count>length) count=length;
		uint32_t word=0;
		memcpy(offset+(char*)&word,data,count);
		acc+=ntohl(word);
		data+=count;
		length-=count;
	}

	char* data_end=data+(length&~3);
	while (data!=data_end) {
		uint32_t word;
		memcpy(&word,data,4);
		acc+=ntohl(word);
		data+=4;
	}
	length&=3;

	if (length) {
		uint32_t word=0;
		memcpy(&word,data,length);
		acc+=ntohl(word);
	}

	acc=(acc&0xffffffff)+(acc>>32);
	while (acc>>16) {
		acc=(acc&0xffff)+(acc>>16);
	}

	if (offset&1) {
		acc=((acc&0xff00)>>8)|((acc&0x00ff)<<8);
	}

	return htons(~acc);
}
