#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <stdio.h>
#include <inttypes.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include "ts_monitor.h"

#define TIMESTAMP_TO_NS(x) \
	((((uint64_t)x.seconds_msb) << 32 | x.seconds_lsb) * 1000000000ULL) + x.nanoseconds

static int running = 1;

key_t tsdata_shmkey = 0xFEE1;
key_t minvio_shmkey = 0xFEE2;
key_t maxvio_shmkey = 0xFEE3;
key_t mpdvio_shmkey = 0xFEE4;
key_t tsent_shmkey = 0xFEE5;

static void handle_int_quit_term(int s)
{
    running = 0;
}

int handle_term_signals(void)
{
    if (SIG_ERR == signal(SIGINT, handle_int_quit_term)) {
            fprintf(stderr, "cannot handle SIGINT\n");
            return -1;
    }
    if (SIG_ERR == signal(SIGQUIT, handle_int_quit_term)) {
            fprintf(stderr, "cannot handle SIGQUIT\n");
            return -1;
    }
    if (SIG_ERR == signal(SIGTERM, handle_int_quit_term)) {
            fprintf(stderr, "cannot handle SIGTERM\n");
            return -1;
    }
    if (SIG_ERR == signal(SIGHUP, handle_int_quit_term)) {
            fprintf(stderr, "cannot handle SIGHUP\n");
            return -1;
    }
    return 0;
}

int main(int argc, char *argv[])
{
    struct time_stamp_entry *offset_min_vio_log;
    struct time_stamp_entry *offset_max_vio_log;
    struct time_stamp_entry *mpd_vio_log;
    struct time_stamp_entry *entry;
    struct time_stamp_data *tsdata;
    struct option opts;
    int tsdata_shmid;
    int minvio_shmid;
    int maxvio_shmid;
    int mpdvio_shmid;
    int tsent_shmid;
    struct tm ntm;
    time_t ntime;
    char buf[32];
    uint16_t idx, logidx;
    int index, i;
    int opt;

    handle_term_signals();

    opt= getopt_long(argc, argv, "oxvlcg", &opts, &index);
    switch (opt) {
        case 'o':
            fprintf(stdout, "Time (UTC)               Seq ID Offset (ns)           MPD (ns)\n");
            fprintf(stdout, "------------------------ ------ -------------------- --------------------\n");
            break;
        case 'x':
            fprintf(stdout, "Time (UTC)               Seq ID Offset (ns)          MPD (ns)             T1 (ns)              T2 (ns)              T3 (ns)              T4 (ns)\n");
            fprintf(stdout, "------------------------ ------ -------------------- -------------------- -------------------- -------------------- -------------------- --------------------\n");
            break;
        case 'v':
        case 'l':
        case 'c':
        case 'g':
            break;
        default:
            fprintf(stderr,
                "\nusage: tsmonitor [options]\n\n"
                " -o    Master Offset information\n"
                " -x    Extended Master Offset information\n"
                " -v    Threshold violations\n"
	        " -l    Threshold violations log \n"\
	        " -c    Clear Threshold violation counters \n"
	        " -g    Clear Threshold violations log \n\n");
            return -1;
    }

    if ((tsdata_shmid = shmget(tsdata_shmkey, sizeof(struct time_stamp_data), 0666)) < 0) {
        perror("shmget time_stamp_data");
        return -1;
    }

    if ((tsdata = shmat(tsdata_shmid, NULL, 0)) == (void *) -1) {
        perror("shmat time_stamp_data");
        return -1;
    }

    if ((minvio_shmid = shmget(minvio_shmkey, tsdata->num_log_sets *
                                    tsdata->num_log_entries *
                                    sizeof(struct time_stamp_entry),
                                    IPC_CREAT | 0666)) < 0) {
        printf("shmget min_vio_log");
        return -1;
    }

    if ((offset_min_vio_log = shmat(minvio_shmid, NULL, 0))
                    == (void *) -1) {
        printf("shmat min_vio_log");
        return -1;
    }

    if ((maxvio_shmid = shmget(maxvio_shmkey, tsdata->num_log_sets *
                                    tsdata->num_log_entries *
                                    sizeof(struct time_stamp_entry),
                                    IPC_CREAT | 0666)) < 0) {
        printf("shmget max_vio_log");
        return -1;
    }

    if ((offset_max_vio_log = shmat(maxvio_shmid, NULL, 0))
                    == (void *) -1) {
        printf("shmat max_vio_log");
        return -1;
    }

    if ((mpdvio_shmid = shmget(mpdvio_shmkey, tsdata->num_log_sets *
                                    tsdata->num_log_entries *
                                    sizeof(struct time_stamp_entry),
                                    IPC_CREAT | 0666)) < 0) {
        printf("shmget mpd_vio_log");
        return -1;
    }

    if ((mpd_vio_log = shmat(mpdvio_shmid, NULL, 0))
                                == (void *) -1) {
        printf("shmat mpd_vio_log");
        return -1;
    }

    if ((tsent_shmid = shmget(tsent_shmkey,
                                    tsdata->num_ts * sizeof(struct time_stamp_entry),
                                    IPC_CREAT | 0666)) < 0) {
        printf("shmget ts_entry");
        return -1;
    }

    if ((entry = shmat(tsent_shmid, NULL, 0))
                                == (void *) -1) {
        printf("shmat ts entry");
        return -1;
    }

    if (opt == 'v') {
	fprintf(stdout,
	        "Maximum number of Time Stamp Entries       : %d\n"
	        "Maximum number of Violation log sets       : %d\n"
	        "Maximum number of Violation entries per set: %d\n"
	        "Violation log wait interval (sec)          : %d\n\n"
		"Minimum Offset Threshold                   : %d\n"
		"Maximum Offset Threshold                   : %d\n"
		"Mean Path Delay Threshold                  : %d\n\n",
		tsdata->num_ts, tsdata->num_log_sets, tsdata->num_log_entries,
			tsdata->log_wait_secs, tsdata->offset_min_threshold,
			tsdata->offset_max_threshold, tsdata->mpd_threshold);

        if (tsdata->offset_min_violations) {
		ntime = tsdata->last_offset_min_violation.tv_sec;
		localtime_r(&ntime, &ntm);
		strftime(buf, sizeof(buf), "%c", &ntm);
	}
	fprintf(stdout,
	        "Minimum Offset Threshold Violations Count  : %6d Last Occurrence: %s\n",
		tsdata->offset_min_violations,
		tsdata->offset_min_violations ? buf : " ");

	if (tsdata->offset_max_violations) {
		ntime = tsdata->last_offset_max_violation.tv_sec;
		localtime_r(&ntime, &ntm);
		strftime(buf, sizeof(buf), "%c", &ntm);
	}

	fprintf(stdout,
		"Maximum Offset Threshold Violations Count  : %6d Last Occurrence: %s\n",
		tsdata->offset_max_violations,
		tsdata->offset_max_violations ? buf : " ");

	if (tsdata->mpd_violations) {
		ntime = tsdata->last_mpd_violation.tv_sec;
		localtime_r(&ntime, &ntm);
		strftime(buf, sizeof(buf), "%c", &ntm);
	}
	fprintf(stdout,
		"Mean Path Delay Threshold Violations Count : %6d Last Occurrence: %s\n",
		tsdata->mpd_violations,
		tsdata->mpd_violations ? buf : " ");

	return 0;
    } else if (opt == 'l') {
        fprintf(stdout, "Time (UTC)               Seq ID Offset (ns)          MPD (ns)             T1 (ns)              T2 (ns)              T3 (ns)              T4 (ns)\n");
        fprintf(stdout, "------------------------ ------ -------------------- -------------------- -------------------- -------------------- -------------------- --------------------\n\n");

		if (tsdata->offset_min_vio_log_index) {
			for (idx = 0, logidx = 0; (idx < tsdata->offset_min_vio_log_index) &&
						(idx < tsdata->num_log_sets); ++idx) {
				fprintf(stdout, "\nMinimum Offset violations log [%d]:\n\n", idx);
				for (i = 0; i < tsdata->num_log_entries; ++i, ++logidx) {
					if (!offset_min_vio_log[logidx].sync_seqid &&
						!offset_min_vio_log[logidx].master_offset.ns &&
						!offset_min_vio_log[logidx].path_delay.ns) {
						continue;
					}
					ntime = offset_min_vio_log[logidx].sync_time.tv_sec;
					localtime_r(&ntime, &ntm);
					strftime(buf, sizeof(buf), "%c", &ntm);
					fprintf(stdout,
				"%24s %6d %20" PRId64" %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 "\n" ,
						buf, offset_min_vio_log[logidx].sync_seqid,
						offset_min_vio_log[logidx].master_offset.ns,
						offset_min_vio_log[logidx].path_delay.ns,
						offset_min_vio_log[logidx].t1.ns,
						offset_min_vio_log[logidx].t2.ns,
						offset_min_vio_log[logidx].t3.ns,
						offset_min_vio_log[logidx].t4.ns);
			  	}
		    }
	    }
	    if (tsdata->offset_max_vio_log_index) {
		    for (idx = 0, logidx = 0; (idx < tsdata->offset_max_vio_log_index) &&
				  (idx < tsdata->num_log_sets); ++idx) {
		        fprintf(stdout, "\nMaximum Offset violations log [%d]:\n\n", idx);
				for (i = 0; i < tsdata->num_log_entries; ++i, ++logidx) {
					if (!offset_max_vio_log[logidx].sync_seqid &&
						!offset_max_vio_log[logidx].master_offset.ns &&
						!offset_max_vio_log[logidx].path_delay.ns) {
						continue;
					}
					ntime = offset_max_vio_log[logidx].sync_time.tv_sec;
					localtime_r(&ntime, &ntm);
					strftime(buf, sizeof(buf), "%c", &ntm);

					fprintf(stdout,
				"%24s %6d %20" PRId64" %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 "\n" ,
						buf, offset_max_vio_log[logidx].sync_seqid,
						offset_max_vio_log[logidx].master_offset.ns,
						offset_max_vio_log[logidx].path_delay.ns,
						offset_max_vio_log[logidx].t1.ns,
						offset_max_vio_log[logidx].t2.ns,
						offset_max_vio_log[logidx].t3.ns,
						offset_max_vio_log[logidx].t4.ns);
				}
		    }
	    }
	    if (tsdata->mpd_vio_log_index) {
		    for (idx = 0, logidx = 0; (idx < tsdata->mpd_vio_log_index) &&
				  (idx < tsdata->num_log_sets); ++idx) {
		        fprintf(stdout, "\nMean Path Delay violations log [%d]:\n\n", idx);
				for (i = 0; i < tsdata->num_log_entries; ++i, ++logidx) {
					if (!mpd_vio_log[logidx].sync_seqid &&
						!mpd_vio_log[logidx].master_offset.ns &&
						!mpd_vio_log[logidx].path_delay.ns) {
						continue;
					}
					ntime = mpd_vio_log[logidx].sync_time.tv_sec;
					localtime_r(&ntime, &ntm);
					strftime(buf, sizeof(buf), "%c", &ntm);

					fprintf(stdout,
				"%24s %6d %20" PRId64" %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 "\n" ,
						buf, mpd_vio_log[logidx].sync_seqid,
						mpd_vio_log[logidx].master_offset.ns,
						mpd_vio_log[logidx].path_delay.ns,
						mpd_vio_log[logidx].t1.ns,
						mpd_vio_log[logidx].t2.ns,
						mpd_vio_log[logidx].t3.ns,
						mpd_vio_log[logidx].t4.ns);
				}
		    }
	    }
	    return 0;
    } else if (opt == 'c') {
	    memset(&tsdata->last_offset_min_violation, 0, sizeof(struct timeval));
	    memset(&tsdata->last_offset_max_violation, 0, sizeof(struct timeval));
	    memset(&tsdata->last_mpd_violation, 0, sizeof(struct timeval));

	    tsdata->offset_min_violations = 0;
	    tsdata->offset_max_violations = 0;
	    tsdata->mpd_violations = 0;

	    return 0;
    } else if (opt == 'g') {
	    tsdata->offset_min_vio_log_index = 0;
	    tsdata->offset_max_vio_log_index = 0;
	    tsdata->mpd_vio_log_index = 0;

	    return 0;
    }

    idx = (tsdata->wr_idx == (tsdata->num_ts - 1)) ? 0 : tsdata->wr_idx + 1;
    
    if (!entry[idx].sync_seqid &&
		!entry[idx].master_offset.ns &&
		!entry[idx].path_delay.ns && idx && (idx < tsdata->num_ts)) {
		idx = 0;
    }

    while (1) {
		if (!entry[idx].sync_seqid &&
			!entry[idx].master_offset.ns &&
			!entry[idx].path_delay.ns) {
			goto Incr_index;
		}

		if (opt == 'o') {
			ntime = entry[idx].sync_time.tv_sec;
			localtime_r(&ntime, &ntm);
			strftime(buf, sizeof(buf), "%c", &ntm);
			fprintf(stdout, "%24s %6d %20" PRId64" %20" PRId64 "\n",
                buf, entry[idx].sync_seqid, entry[idx].master_offset.ns,
                entry[idx].path_delay.ns);
		} else if (opt == 'x') {
			ntime = entry[idx].sync_time.tv_sec;
			localtime_r(&ntime, &ntm);
			strftime(buf, sizeof(buf), "%c", &ntm);
			fprintf(stdout, "%24s %6d %20" PRId64" %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 " %20" PRId64 "\n",
				buf, entry[idx].sync_seqid, entry[idx].master_offset.ns,
				entry[idx].path_delay.ns, entry[idx].t1.ns,
				entry[idx].t2.ns, entry[idx].t3.ns,
				entry[idx].t4.ns);
		}
Incr_index:
		if (++idx == tsdata->num_ts) {
			idx = 0;
		}
		if (idx == tsdata->wr_idx) {
			break;
		}
    }

}
